from typing import overload, Union, Sequence, Tuple, Callable, Optional
import operator
import math
from numbers import Number
from functools import partial, reduce
from itertools import product, accumulate
from cachetools import LRUCache
import torch
from torch import Tensor
import torch.nn.functional as F
from torch.autograd import Function
from torch.autograd.function import FunctionCtx
import torch.fx
from ..utils import compilable
from ..utils.devices import multidevices


@overload
def unfold(input: Tensor, dimension: int, size: int, step: int) -> Tensor: ...
unfold = Tensor.unfold


def fold_roll(input: Tensor, dimension: int, step: int) -> Tensor:
    return _fold_roll_overlap_implementation[input.device.type](input, dimension, step)


def fold(input: Tensor, dimension: int, step: int) -> Tensor:
    return _Fold.apply(input, dimension, step)


def unfold_space(
    input: Tensor,
    kernel_size: Union[int, Sequence[int]],
    stride: Union[int, Sequence[int]],
    padding: Union[int, Sequence[int]] = 0,
    dilation: Union[int, Sequence[int]] = 1, # cannot be recovered through `fold_space`
    ceil_mode: bool = False, # cannot be recovered through `fold_space`
    *,
    default_space: Union[float, str] = 0
) -> Tensor:
    if isinstance(default_space, Number):
        functional_pad = partial(F.pad, mode='constant', value=default_space)
    else:
        functional_pad = partial(F.pad, mode=default_space)
    
    spatial_ndim = input.ndim - 2
    
    kernel_size = _spatialize_tuple(kernel_size, spatial_ndim)
    stride = _spatialize_tuple(stride, spatial_ndim)
    
    while padding != 0:
        if isinstance(padding, int):
            padding = (padding,) * (spatial_ndim * 2)
        elif any(value != 0 for value in padding):
            padding = [value for value in padding[::-1] for _ in range(2)]
        else:
            break
        input = functional_pad(input, padding)
        break
    
    if ceil_mode:
        spatial_shape = input.shape[2:]
        ceil_pad = (math.ceil((length - size) / step) * step + size - length for length, size, step in zip(reversed(spatial_shape), reversed(kernel_size), reversed(stride)))
        if any(pad != 0 for pad in ceil_pad):
            input = functional_pad(input, [pad_value for value in ceil_pad for pad_value in (0, value)])
    
    if dilation == 1:
        iterator = zip(kernel_size, stride)
        single_unfold = unfold
    else:
        dilation = _spatialize_tuple(dilation, spatial_ndim)
        iterator = zip(kernel_size, stride, dilation)
        single_unfold = _dilatedly_unfold
    
    output = input
    for dimension, unfold_args in enumerate(iterator, 2):
        output = single_unfold(output, dimension, *unfold_args)
    return output


def fold_space(
    input: Tensor,
    stride: Union[int, Sequence[int]],
    padding: Union[int, Sequence[int]] = 0,
    dilation: Union[int, Sequence[int]] = 1
) -> Tensor:
    return _fold_space(input, stride, padding, dilation, folder=_FoldSpace.apply)


def fold_stack(
    input: Tensor,
    stride: Union[int, Sequence[int]],
    padding: Union[int, Sequence[int]] = 0,
    dilation: Union[int, Sequence[int]] = 1
) -> Tensor:
    return _fold_space(input, stride, padding, dilation, folder=_fold_stack_overlap)


def patches_column(input: Tensor) -> Tensor:
    """
    Args:
        input (Tensor): [B, C, out_L_0, out_L_1, ..., out_L_n, k_L_0, k_L_1, ..., k_L_n]
    
    Returns:
        output (Tensor): [B, C * k_L_0 * k_L_1 * ... * k_L_n, out_L_0 * out_L_1 * ... * out_L_n]
    
    """
    double_spatial_ndim = input.ndim - 2
    if double_spatial_ndim % 2 != 0:
        raise ValueError(f"'input' is expected to be a tensor that was completely spatially unfolded")
    spatial_ndim = double_spatial_ndim // 2
    
    output = input.permute(0, 1, *range(2 + spatial_ndim, input.ndim), *range(2, 2 + spatial_ndim))
    output_shape = output.shape
    return output.reshape(output_shape[0], -1, math.prod(output_shape[-spatial_ndim:]))


def keep_kernel_dim(input: Tensor) -> Tensor:
    spatial_ndim = input.ndim - 2
    return input.view(*input.shape, *((1,) * spatial_ndim))


def expand_kernel_dim(input: Tensor, kernel_size: Sequence[int]) -> Tensor:
    spatial_ndim = input.ndim - 2
    output = input.view(*input.shape, *((1,) * spatial_ndim))
    if spatial_ndim != len(kernel_size):
        raise ValueError(f"the length of 'kernel_size' should be equal to the number of spatial dimensions of 'input'")
    output = output.expand(*((-1,) * input.ndim), *kernel_size)
    return output


def fold2d_roll(input: Tensor, stride: Tuple[int, int]) -> Tensor:
    device = input.device
    
    input_shape = input.shape
    batch_size, num_channels = input_shape[:2]
    other_shape = torch.as_tensor(input_shape[2:], device=device)
    kernel_size = other_shape[-2:]
    output_size = (other_shape[:2] - 1) * torch.as_tensor(stride, device=device) + kernel_size
    functional_fold_input = input.permute(0, 1, 4, 5, 2, 3).reshape(batch_size, num_channels * torch.prod(kernel_size).item(), -1)
    
    output = F.fold(
        input=functional_fold_input,
        output_size=output_size,
        kernel_size=kernel_size,
        stride=stride
    )
    
    return output


class _FoldRoll(Function):
    @classmethod
    def initialize(cls, ctx: FunctionCtx, input: Tensor, dimension: int, step: int) -> Tensor:
        ctx.dimension = dimension
        ctx.step = step
        output_shape = list(input.shape[:-1])
        num_patches = output_shape[dimension]
        size = input.size(-1)
        ctx.size = size
        output_shape[dimension] = (num_patches - 1) * step + size
        return output_shape
    
    @staticmethod
    def backward(ctx: FunctionCtx, grad_output: Optional[Tensor]):
        if grad_output is None:
            return None, None, None
        
        dimension = ctx.dimension
        size = ctx.size
        step = ctx.step
        
        grad_input = grad_output.unfold(dimension, size, step)
        return grad_input, None, None


class _CheckerboardFoldRoll(_FoldRoll):
    @staticmethod
    def forward(ctx: FunctionCtx, input: Tensor, dimension: int, step: int) -> Tensor:
        output_shape = __class__.initialize(ctx, input, dimension, step)
        size = ctx.size
        
        # Checkerboard Algorithm
        # Sometimes faster than `F.fold` when `step` > 1
        output: Tensor = torch.zeros(output_shape, device=input.device, dtype=input.dtype)
        patch = output.unfold(dimension, size, step)

        # Checkerboard Partitioning
        # Calculate security step size, since as long as one block is taken every step, the extracted blocks will not physically overlap with each other.
        # `step = ceil(K / S)`
        # Generate all offset combinations (Offsets)
        offset_range = range(math.ceil(size / step))
        
        # Pre-defined full slice: `(slice(None), slice(None))` -> Batch, Channel
        base_slices = (slice(None),) * dimension
        for offset in offset_range:
            # Construct the slice object for the current iteration.
            full_slice = (*base_slices, slice(offset, None, step))
            
            # Due to the large step size, any two elements in that point to physical memory addresses that do not overlap.
            # Therefore, it can be safely do Inplace Add, which is extremely efficient.
            patch[full_slice].add_(input[full_slice])

        return output
_fold_roll_overlap_implementation = multidevices(_CheckerboardFoldRoll.apply)


def _create_term(
    graph: torch.fx.Graph,
    shape_node: torch.fx.Node,
    device_node: torch.fx.Node,
    index: int,
    factor: int,
    input_ndim: int
) -> torch.fx.Node:
    dim_size = graph.call_function(operator.getitem, (shape_node, index))
    view_shape = [1] * input_ndim
    view_shape[index] = dim_size
    arange = graph.call_function(torch.arange, (dim_size,), {"device": device_node})
    view = graph.call_method("view", (arange, view_shape))
    return graph.call_function(operator.mul, (view, factor))


# The `indices` may occupy a large amount of memory, but it may be eliminated by `torch.compile` when this is supported.
def _fold_roll_overlap_indexing_graph(dimension: int, step: int, input_ndim: int) -> torch.fx.GraphModule:
    graph = torch.fx.Graph()
    
    input_node = graph.placeholder("input")
    shape_node = graph.call_function(getattr, (input_node, "shape"))
    shape_nodes = [graph.call_function(operator.getitem, (shape_node, i)) for i in range(input_ndim)]
    
    num_patches_node = shape_nodes[dimension]
    kernel_size_node = shape_nodes[-1]
    
    # Output Spatial Size: (L - 1) * step + K
    output_spatial_node = graph.call_function(operator.add, (
        graph.call_function(operator.mul, (
            graph.call_function(operator.sub, (num_patches_node, 1)), 
            step
        )),
        kernel_size_node
    ))
    
    output_shape_nodes = shape_nodes[:-1]
    output_shape_nodes[dimension] = output_spatial_node
    output_stride = list(accumulate(output_shape_nodes[:0:-1], lambda left, right: graph.call_function(operator.mul, (left, right)), initial=1))[::-1]
    
    device_node = graph.call_function(getattr, (input_node, "device"))
    dtype_node = graph.call_function(getattr, (input_node, "dtype"))
    
    output_node = graph.call_function(torch.zeros, (output_shape_nodes,), {"device": device_node, "dtype": dtype_node})
    
    # Calculate linear factor
    indices_factor = (
        *output_stride[:dimension],
        graph.call_function(operator.mul, (output_stride[dimension], step)),
        *output_stride[dimension+1:],
        output_stride[dimension]
    )
    
    indices_node = reduce(
        lambda left, right: graph.call_function(operator.add, (left, right)),
        map(partial(_create_term, graph, shape_node, device_node, input_ndim=input_ndim), range(input_ndim), indices_factor)
    )
    
    graph.call_method("put_", (
        graph.call_method("flatten", (output_node,)),
        indices_node,
        input_node
    ), {"accumulate": True})
    
    graph.output(output_node)
    return torch.fx.GraphModule({}, graph)


class _IndexingFoldRoll(_FoldRoll):
    _cache = LRUCache(maxsize=compilable.recompile_limit())
    @staticmethod
    def forward(ctx: FunctionCtx, input: Tensor, dimension: int, step: int) -> Tensor:
        input_ndim = input.ndim
        
        key = (dimension, step, input_ndim)
        cache = __class__._cache
        try:
            compiled_function = cache[key]
        except:
            compiled_function = compilable.compile(_fold_roll_overlap_indexing_graph(dimension, step, input_ndim), fullgraph=True)
            cache[key] = compiled_function
        output: Tensor = compiled_function(input)
        
        ctx.dimension = dimension
        ctx.size = input.size(-1)
        ctx.step = step
        
        return output
for device in compilable.compilable_gpus(): _fold_roll_overlap_implementation.register_device(device, _IndexingFoldRoll.apply)


class _CpuFoldRoll(_FoldRoll):
    @staticmethod
    def forward(ctx: FunctionCtx, input: Tensor, dimension: int, step: int) -> Tensor:
        output_shape = __class__.initialize(ctx, input, dimension, step)
        
        output: Tensor = torch.zeros(output_shape, device=input.device, dtype=input.dtype)
        patch = output.unfold(dimension, ctx.size, step)
        patch.add_(input)
        return output
_fold_roll_overlap_implementation.register_device('cpu', _CpuFoldRoll.apply)


class _Fold(Function):
    @staticmethod
    def forward(ctx: FunctionCtx, input: Tensor, dimension: int, step: int) -> Tensor:
        device = input.device
        input_ndim = input.ndim

        implementation = _fold_roll_overlap_implementation[device.type]
        expansion = [1] * input_ndim
        ones = torch.ones(expansion, device=device, dtype=input.dtype)
        expansion[dimension] = input.size(dimension)
        expansion[-1] = input.size(-1)
        ones = ones.expand(expansion)
        count: Tensor = implementation(
            ones, dimension, step
        )
        count.clamp_min_(1)
        output: Tensor = implementation(input, dimension, step)
        output.div_(count)
        
        ctx.save_for_backward(count)
        ctx.dimension = dimension
        ctx.size = input.size(-1)
        ctx.step = step
        
        return output
    
    @staticmethod
    def backward(ctx: FunctionCtx, grad_output: Optional[Tensor]):
        if grad_output is None:
            return None, None, None
        
        count: Tensor = ctx.saved_tensors[0]
        dimension = ctx.dimension
        size = ctx.size
        step = ctx.step
        
        grad_input = (grad_output / count).unfold(dimension, size, step)
        return grad_input, None, None


def _fold_space(
    input: Tensor,
    stride: Union[int, Sequence[int]],
    padding: Union[int, Sequence[int]] = 0,
    dilation: Union[int, Sequence[int]] = 1,
    *,
    folder: Callable[[Tensor, Sequence[int]], Tensor]
) -> Tensor:
    double_spatial_ndim = input.ndim - 2
    if double_spatial_ndim % 2 != 0:
        raise ValueError(f"'input' is expected to be a tensor that was completely spatially unfolded")
    spatial_ndim = double_spatial_ndim // 2
    
    while dilation != 1:
        if isinstance(dilation, int):
            dilation = _spatialize_tuple(dilation, spatial_ndim)
        elif not any(value != 1 for value in dilation):
            break
        input = _dilate_patch(input, dilation, spatial_ndim)
        break
    
    stride = _spatialize_tuple(stride, spatial_ndim)
    
    output = folder(input, stride)
    
    if padding != 0:
        padding = _spatialize_tuple(padding, spatial_ndim)
        output = output[(..., *(slice(pad, -pad) for pad in padding))]
    
    return output


def _dilatedly_unfold(input: Tensor, dimension: int, size: int, step: int, dilation: int) -> Tensor:
    if dilation == 1:
        return unfold(input, dimension, size, step)
    effective_size = (size - 1) * dilation + 1
    return input.unfold(dimension, effective_size, step)[..., ::dilation]


def _dilate_patch(input: Tensor, dilation: Sequence[int], spatial_ndim: int) -> Tensor:
    shape = input.shape
    num_patches = shape[2:-spatial_ndim]
    patch_size =  shape[-spatial_ndim:]
    effective_patch_size = [(each_patch_size - 1) * each_dilation + 1 for each_patch_size, each_dilation in zip(patch_size, dilation)]
    def generate_index():
        dilation_length = len(dilation)
        for i, (size, expansion) in enumerate(zip(num_patches, dilation)):
            yield torch.arange(size, device=input.device).mul_(expansion)[(slice(None), *((None,) * (dilation_length - i - 1)))]
    expanded = input.new_zeros(*shape[:-spatial_ndim], *effective_patch_size)
    expanded[(..., *generate_index())] = input
    return expanded


def _spatialize_tuple(parameter: Union[int, Sequence[int]], spatial_ndim: int) -> tuple[int, ...]:
    if isinstance(parameter, int):
        parameter = (parameter,) * spatial_ndim
    else:
        parameter_length = len(parameter)
        if parameter_length < spatial_ndim:
            parameter = (*((1,) * spatial_ndim - parameter_length), *parameter)
    return parameter


class _FoldStack(Function):
    @classmethod
    def initialize(cls, ctx: FunctionCtx, input: Tensor, stride: Sequence[int]):
        double_spatial_ndim = input.ndim - 2
        if double_spatial_ndim % 2 != 0:
            raise ValueError(f"'input' is expected to be a tensor that was completely spatially unfolded")
        spatial_ndim = double_spatial_ndim // 2

        batch_size, num_channels, *other_shape = input.shape
        kernel_size = other_shape[-spatial_ndim:]
        output_shape = (batch_size, num_channels, *(
            (num_patches - 1) * step + size
            for num_patches, size, step in zip(other_shape[:-spatial_ndim], kernel_size, stride)
        ))
        ctx.kernel_size = kernel_size
        ctx.stride = stride
        return output_shape
    
    @staticmethod
    def backward(ctx: FunctionCtx, grad_output: Optional[Tensor]):
        if grad_output is None:
            return None, None
        
        kernel_size = ctx.kernel_size
        stride = ctx.stride
        
        grad_input = grad_output
        for dimension, (size, step) in enumerate(zip(kernel_size, stride), 2):
            grad_input = grad_input.unfold(dimension, size, step)
        return grad_input, None


class _CheckerboardFoldStack(_FoldStack):
    @staticmethod
    def forward(ctx: FunctionCtx, input: Tensor, stride: Sequence[int]) -> Tensor:
        output_shape = __class__.initialize(ctx, input, stride)
        kernel_size = ctx.kernel_size
        
        # Checkerboard Algorithm
        # Sometimes faster than `F.fold` when `stride` > 1
        output: Tensor = torch.zeros(output_shape, device=input.device, dtype=input.dtype)
        patch = output
        for dimension, (size, step) in enumerate(zip(kernel_size, stride), 2):
            patch = patch.unfold(dimension, size, step)

        # Checkerboard Partitioning
        # Calculate security step size, since as long as one block is taken every step, the extracted blocks will not physically overlap with each other.
        # `step = ceil(K / S)`
        steps = [math.ceil(size / step) for size, step in zip(kernel_size, stride)]
        
        # Generate all offset combinations (Offsets)
        offset_ranges = (range(step) for step in steps)
        
        # Pre-defined full slice: `(slice(None), slice(None))` -> Batch, Channel
        base_slices = (slice(None),) * 2
        for offsets in product(*offset_ranges):
            # Construct the slice object for the current iteration.
            full_slice = (*base_slices, *(slice(offset, None, step) for offset, step in zip(offsets, steps)))
            
            # Due to the large step size, any two elements in that point to physical memory addresses that do not overlap.
            # Therefore, it can be safely do Inplace Add, which is extremely efficient.
            patch[full_slice].add_(input[full_slice])

        return output
_fold_stack_overlap_implementation = multidevices(_CheckerboardFoldStack.apply)


# The `indices` may occupy a large amount of memory, but it may be eliminated by `torch.compile` when this is supported.
def _fold_stack_overlap_indexing_graph(stride: Sequence[int], input_ndim: int) -> torch.fx.GraphModule:
    graph = torch.fx.Graph()
    
    input_node = graph.placeholder("input")
    shape_node = graph.call_function(getattr, (input_node, "shape"))
    shape_nodes = [graph.call_function(operator.getitem, (shape_node, i)) for i in range(input_ndim)]
    
    spatial_ndim = (input_ndim - 2) // 2
    watershed_dimension = 2 + spatial_ndim
    
    num_patches_nodes = shape_nodes[2:watershed_dimension]
    kernel_size_nodes = shape_nodes[watershed_dimension:]

    output_spatial_nodes = [
        graph.call_function(operator.add, (graph.call_function(operator.mul, (graph.call_function(operator.sub, (num_patches_node, 1)), step)), size_node))
        for num_patches_node, size_node, step in zip(num_patches_nodes, kernel_size_nodes, stride)
    ]
    output_shape = shape_nodes[:2] + output_spatial_nodes
    output_stride = list(accumulate(output_shape[:0:-1], lambda left, right: graph.call_function(operator.mul, (left, right)), initial=1))[::-1]
    
    device_node = graph.call_function(getattr, (input_node, "device"))
    dtype_node = graph.call_function(getattr, (input_node, "dtype"))
    
    output_node = graph.call_function(torch.zeros, (output_shape,), {"device": device_node, "dtype": dtype_node})
    
    # Calculate linear factor
    indices_factor = (
        *output_stride[:2], 
        *(graph.call_function(operator.mul, (output_step, step)) for output_step, step in zip(output_stride[2:], stride)), 
        *output_stride[2:]
    )
    indices_node = reduce(
        lambda left, right: graph.call_function(operator.add, (left, right)),
        map(partial(_create_term, graph, shape_node, device_node, input_ndim=input_ndim), range(input_ndim), indices_factor)
    )
    
    graph.call_method("put_", (
        graph.call_method("flatten", (output_node,)),
        indices_node,
        input_node
    ), {"accumulate": True})
    
    graph.output(output_node)
    return torch.fx.GraphModule({}, graph)


class _IndexingFoldStack(_FoldStack):
    _cache = LRUCache(maxsize=compilable.recompile_limit())
    @staticmethod
    def forward(ctx: FunctionCtx, input: Tensor, stride: Sequence[int]) -> Tensor:
        input_ndim = input.ndim
        
        key = (*stride, input_ndim)
        cache = __class__._cache
        try:
            compiled_function = cache[key]
        except:
            compiled_function = compilable.compile(_fold_stack_overlap_indexing_graph(stride, input_ndim), fullgraph=True)
            cache[key] = compiled_function
        output: Tensor = compiled_function(input)
        
        ctx.kernel_size = input.shape[output.ndim - input_ndim:]
        ctx.stride = stride

        return output
for device in compilable.compilable_gpus(): _fold_stack_overlap_implementation.register_device(device, _IndexingFoldStack.apply)


class _CpuFoldStack(_FoldStack):
    @staticmethod
    def forward(ctx: FunctionCtx, input: Tensor, stride: Sequence[int]) -> Tensor:
        output_shape = __class__.initialize(ctx, input, stride)

        output: Tensor = torch.zeros(output_shape, device=input.device, dtype=input.dtype)
        patch = output
        for dimension, (size, step) in enumerate(zip(ctx.kernel_size, stride), 2):
            patch = patch.unfold(dimension, size, step)
        patch.add_(input)
        return output
_fold_stack_overlap_implementation.register_device('cpu', _CpuFoldStack.apply)


def _fold_stack_overlap(input: Tensor, stride: Sequence[int]) -> Tensor:
    """
    Args:
        input (Tensor): [B, C, out_L_0, out_L_1, ..., out_L_n, k_L_0, k_L_1, ..., k_L_n]
    
    """
    return _fold_stack_overlap_implementation[input.device.type](input, stride)


class _FoldSpace(Function):
    @staticmethod
    def forward(ctx: FunctionCtx, input: Tensor, stride: Sequence[int]) -> Tensor:
        device = input.device
        spatial_shape_with_kernel_size = input.shape[2:]
        implementation = _fold_stack_overlap_implementation[device.type]
        count: Tensor = implementation(
            torch.ones(
                (1,) * (2 + len(spatial_shape_with_kernel_size)),
                device=device,
                dtype=input.dtype
            ).expand(1, 1, *spatial_shape_with_kernel_size),
            stride
        )
        count.clamp_min_(1)
        
        output: Tensor = implementation(input, stride)
        output.div_(count)
        
        ctx.save_for_backward(count)
        ctx.kernel_size = input.shape[output.ndim - input.ndim:]
        ctx.stride = stride
        
        return output
    
    @staticmethod
    def backward(ctx: FunctionCtx, grad_output: Optional[Tensor]):
        if grad_output is None:
            return None, None
        
        count, = ctx.saved_tensors
        kernel_size = ctx.kernel_size
        stride = ctx.stride
        
        grad_input: Tensor = grad_output / count
        for dimension, (size, step) in enumerate(zip(kernel_size, stride), 2):
            grad_input = grad_input.unfold(dimension, size, step)
        return grad_input, None


from itertools import chain
import torch
from .typing import overload, Iterator
from .devices import supported_devices


try:
    from torch import compile
    
    
    try:
        from torch._dynamo import config as dynamo_config
        def recompile_limit() -> int: return dynamo_config.cache_size_limit
    except:
        def recompile_limit() -> int: return 8
    
    
    @compile
    def _empty(device: str) -> torch.Tensor: return torch.empty([], device=device)


    def _check_device_compilable(device: str) -> bool:
        try:
            _empty(device)
            return True
        except:
            return False
    
    
    _is_cpu_compilable = _check_device_compilable("cpu")
    
    
    try:
        from torch._inductor import utils
        _compilable_gpus = set(utils.GPU_TYPES)
    except AttributeError:
        try:
            from torch._dynamo import device_interface as _
            _compilable_gpus = {"cuda", "xpu"}
        except ImportError:
            _compilable_gpus = {"cuda"}
        except Exception as e: raise e from None
    except Exception:
        _compilable_gpus = set(device for device in supported_devices() if device != "cpu" and _check_device_compilable(device))
    
    
    def compilable_gpus() -> Iterator[str]:
        return iter(_compilable_gpus)
    
    
    def compilable_devices() -> Iterator[str]:
        if _is_cpu_compilable:
            return chain(("cpu",), compilable_gpus())
        else:
            return compilable_gpus()
    
    
    def is_device_compilable(device: str) -> bool:
        if device == 'cpu':
            return _is_cpu_compilable
        else:
            return device in _compilable_gpus
except:
    def recompile_limit() -> int: return 0
    
    
    def compilable_gpus() -> Iterator[str]: return (None for _ in range(0))
    
    
    def compilable_devices() -> Iterator[str]: return compilable_gpus()
    
    
    @overload
    def is_device_compilable(device: str) -> bool: return False
    
    
    @overload
    def compile(model = None, **kwargs):
        if model is None:
            def compiler(model): return model
            return compiler
        return model


from typing import Callable, Iterable, Sequence, Optional, Union, Tuple
from functools import reduce
from itertools import chain
import torch
from .nn import Buffer


def apply_from_dim(function: Callable, input: torch.Tensor, dim = 0, otypes: Iterable[torch.dtype] = None) -> Union[torch.Tensor, Sequence[torch.Tensor]]:
    '''
    Select a dim of an `torch.Tensor` and apply a function.
    '''
    slices = (slice(None, None, None),) * dim
    if len(otypes) > 1:
        return tuple(torch.as_tensor(item, dtype=otypes[i], device=input.device) for i, item in enumerate(zip(*[function(input[slices + (i,)]) for i in range(input.shape[dim])])))
    elif len(otypes) == 1:
        return torch.as_tensor([function(input[slices + (i,)]) for i in range(input.shape[dim])], otypes[0], device=input.device)
    else:
        [function(input[slices + (i,)]) for i in range(input.shape[dim])]


min = torch.amin


max = torch.amax


def map_range(
    input: torch.Tensor,
    interval: Sequence[int] = (0, 1),
    dim: Union[int, Sequence[int], None] = None,
    dtype: torch.dtype = None,
    scalar_default: Union[str, None] = 'max',
    eps: float = 1e-6
) -> torch.Tensor:
    min_value: torch.Tensor = torch.amin(input, dim=dim, keepdim=True)
    max_value: torch.Tensor = torch.amax(input, dim=dim, keepdim=True)
    max_min_difference = max_value - min_value
    max_min_equal_mask = max_min_difference == 0
    max_min_difference.masked_fill_(max_min_equal_mask, 1)
    input = input - min_value
    if not (scalar_default is None or scalar_default == 'none'):
        input.masked_fill_(max_min_equal_mask, torch.tensor(_scalar_default_value(scalar_default, eps)).to(input.dtype))
    return (input / max_min_difference * (interval[1] - interval[0]) + interval[0]).to(dtype)


def map_ranges(
    input: torch.Tensor,
    intervals: Sequence[Sequence[int]] = [(0, 1)],
    dim: Union[int, Sequence[int], None] = None,
    dtype: torch.dtype = None,
    scalar_default: Union[str, None] = 'max',
    eps: float = 1e-6
) -> Tuple[torch.Tensor, ...]:
    min_value: torch.Tensor = torch.amin(input, dim=dim, keepdim=True)
    max_value: torch.Tensor = torch.amax(input, dim=dim, keepdim=True)
    max_min_difference = max_value - min_value
    max_min_equal_mask = max_min_difference == 0
    max_min_difference.masked_fill_(max_min_equal_mask, 1)
    input = input - min_value
    if not (scalar_default is None or scalar_default == 'none'):
        input.masked_fill_(max_min_equal_mask, torch.tensor(_scalar_default_value(scalar_default, eps)).to(input.dtype))
    normed = input / max_min_difference
    def generator():
        for interval in intervals:
            yield (normed * (interval[1] - interval[0]) + interval[0]).to(dtype)
    return tuple(*generator())


def gamma(input: torch.Tensor, *, out: torch.Tensor | None = None) -> torch.Tensor:
    return torch.exp(torch.lgamma(input, out=out), out=out)


def gamma_div(left: torch.Tensor, right: torch.Tensor, *, out: torch.Tensor | None = None) -> torch.Tensor:
    return torch.exp(torch.lgamma(left, out=out) - torch.lgamma(right, out=out), out=out)


def recur_lgamma(n: torch.Tensor, base: torch.Tensor):
    out = torch.lgamma(base + n) - torch.lgamma(base)
    gamma_sign = torch.empty_like(base, dtype=torch.int64)
    gamma_sign[base > 0] = 1
    negative_one_exponent = torch.ceil(-base).to(device=base.device, dtype=int)
    negative_one_exponent[negative_one_exponent > n] = torch.broadcast_to(n, negative_one_exponent.shape)[negative_one_exponent > n]
    negative_base_mask = base < 0
    even_mod_exponent = negative_one_exponent % 2
    gamma_sign[negative_base_mask & (even_mod_exponent == 0)] = 1
    gamma_sign[negative_base_mask & (even_mod_exponent != 0)] = -1
    return out, gamma_sign


def arith_gamma_prod(arith_term: torch.Tensor, arith_base: torch.Tensor, ratio_base: torch.Tensor) -> torch.Tensor:
    arith_lgamma, gamma_sign = recur_lgamma(arith_term, arith_base)
    ratio_base_sign = torch.sign(ratio_base)
    ratio_base_sign[(ratio_base_sign == -1) & (arith_term % 2 == 0)] = 1
    return torch.exp(arith_term * torch.log(torch.abs(ratio_base)) + arith_lgamma) * gamma_sign * ratio_base_sign


def linspace(start, stop, num, dtype: Optional[torch.dtype] = None) -> torch.Tensor:
    start = torch.as_tensor(start)
    stop = torch.as_tensor(stop)
    num = torch.as_tensor(num)
    common_difference = torch.as_tensor((stop - start) / (num - 1).to(dtype=dtype))
    index = torch.arange(num).to(common_difference.device)
    return start + common_difference * index.view([*index.shape] + [1] * len(common_difference.shape))


def linspace_at(index, start, stop, num, dtype: Optional[torch.dtype] = None) -> torch.Tensor:
    start = torch.as_tensor(start)
    stop = torch.as_tensor(stop)
    num = torch.as_tensor(num)
    common_difference = torch.as_tensor((stop - start) / (num - 1).to(dtype=dtype))
    index = torch.as_tensor(index)
    return start + common_difference * index.view([*index.shape] + [1] * len(common_difference.shape))


def invert(input: torch.Tensor) -> torch.Tensor:
    shape_length = len(input.shape)
    if shape_length > 2:
        dims = (shape_length - 2, shape_length - 1)
    else:
        dims = (shape_length - 1,)
    min_values = torch.amin(input, dim=dims, keepdim=True)
    max_values = torch.amax(input, dim=dims, keepdim=True)
    return max_values - input + min_values


def buffer(tensor: Optional[torch.Tensor], persistent: bool = True) -> torch.Tensor:
    return Buffer(tensor, persistent)


def advanced_indexing(shape: Sequence[int], indices: torch.Tensor, dim: int):
    len_shape = len(shape)
    indices_shape = indices.shape
    left_args = _limited_dim_indexing(shape[:dim], len_shape)
    right_args = _limited_dim_indexing(shape[dim + 1:], len_shape, dim + 1)
    return *left_args, indices.view((*indices_shape,) + (1,) * (len_shape - len(indices_shape))), *right_args


def grow(input: torch.Tensor, ndim: int, direction: str = 'leading') -> torch.Tensor:
    tensor_shape = input.shape
    remaining_dims_num = ndim - len(tensor_shape)
    if direction == 'leading':
        return input.view((*((1,) * remaining_dims_num), *tensor_shape))
    elif direction == 'trailing':
        return input.view((*tensor_shape, *((1,) * remaining_dims_num)))
    else:
        return input


def is_int(dtype: torch.dtype) -> bool:
    try:
        torch.iinfo(dtype)
        return True
    except TypeError:
        return False


def is_float(dtype: torch.dtype) -> bool:
    return dtype.is_floating_point
    # try:
    #     torch.finfo(dtype)
    #     return True
    # except TypeError:
    #     return False


def _shift_arguments(input: torch.Tensor, shift: torch.Tensor):
    batch_size, num_channels, *spatial_shape = input.shape
    length_ndim = len(spatial_shape)
    expected_shifts_size = torch.Size((batch_size, num_channels, length_ndim))
    
    if is_float(shift.dtype):
        shift = shift.round().int()
    if shift.shape != expected_shifts_size:
        shift = shift.broadcast_to(expected_shifts_size)
    del expected_shifts_size
    
    device = input.device

    aranged = torch.arange(max(input.shape), device=device)
    batch_size_indices, num_channels_indices, *axes = tuple(aranged[:each] for each in input.shape)
    grid = torch.stack(torch.meshgrid(*axes, indexing='ij'))
    del aranged, axes
    grid = grid.unsqueeze(1).unsqueeze(1).expand(-1, batch_size, num_channels, *spatial_shape)

    expanded_shift = shift.permute(2, 0, 1).reshape(length_ndim, batch_size, num_channels, *([1] * length_ndim))
    source_indices = grid - expanded_shift
    del expanded_shift, shift

    length = torch.as_tensor(spatial_shape, device=device).reshape(length_ndim, *([1] * (2 + length_ndim)))
    valid = ((source_indices >= 0) & (source_indices < length)).all(dim=0)
    del length

    batch_stride, channels_stride, *length_stride = input.stride()
    length_stride = torch.as_tensor(length_stride, device=device).reshape(length_ndim, *([1] * (2 + length_ndim)))

    destination_inplane = (grid * length_stride).sum(dim=0)
    source_inplane = (source_indices * length_stride).sum(dim=0)
    del grid, length_stride, source_indices

    batch_size_indices = batch_size_indices.view(batch_size, 1, *([1] * length_ndim)).expand(batch_size, num_channels, *spatial_shape)
    num_channels_indices = num_channels_indices.view(1, num_channels, *([1] * length_ndim)).expand(batch_size, num_channels, *spatial_shape)
    batch_channels_offset = batch_size_indices * batch_stride + num_channels_indices * channels_stride
    del batch_stride, channels_stride, batch_size_indices, num_channels_indices

    source_flat = (batch_channels_offset + source_inplane).reshape(-1)
    destination_flat = (batch_channels_offset + destination_inplane).reshape(-1)
    del destination_inplane, source_inplane, batch_channels_offset
    
    mask = valid.reshape(-1)
    return destination_flat, source_flat, mask


def shift_(input: torch.Tensor, shift: torch.Tensor, fill_value: int = 0) -> torch.Tensor:
    destination_flat, source_flat, mask = _shift_arguments(input, shift)
    input.view(-1)[destination_flat[mask]] = input.view(-1)[source_flat[mask]]
    input.view(-1)[destination_flat[~mask]] = fill_value
    return input


def shift(input: torch.Tensor, shift: torch.Tensor, fill_value: int = 0) -> torch.Tensor:
    out = torch.full_like(input, fill_value)
    destination_flat, source_flat, mask = _shift_arguments(input, shift)
    out.view(-1)[destination_flat[mask]] = input.view(-1)[source_flat[mask]]
    return out


def promote_types(*types) -> torch.dtype:
    return reduce(torch.promote_types, types)


def _scalar_default_value(scalar_default, eps=1e-6):
    if scalar_default == 'max':
        return 1 - eps
    elif scalar_default == 'min':
        return eps
    else:
        return 0.5


def _tailing_dim(ndim: int, dim: Sequence[int]):
    inv_dims = [0] * ndim
    def generate_dims():
        nonlocal inv_dims
        for permute_back_dim, permuted_dim in enumerate(chain((d for d in range(ndim) if d not in dim), dim)):
            inv_dims[permuted_dim] = permute_back_dim
            yield permuted_dim
    tailing_dims = type(dim)(generate_dims())
    return tailing_dims, inv_dims


def _limited_dim_indexing(using_shape, len_shape, start: int = 0):
    return (torch.arange(dim_size).view((1,) * (i + start) + (dim_size,) + (1,) * (len_shape - (i + start) - 1)) for i, dim_size in enumerate(using_shape))


from typing import overload, Iterator, Callable, Any, Optional
import torch


_broad_supported_devices = ['cpu', 'cuda', 'mkldnn', 'opengl', 'opencl', 'ideep', 'hip', 'msnpu', 'xla']
_supported_devices = []
def supported_devices() -> Iterator[str]:
    global _supported_devices
    if not _supported_devices:
        try:
            torch.device("_")
        except RuntimeError as e:
            try:
                msg = e.args[0]
                _supported_devices = [device.strip() for device in msg[16:-41].split(',') if device.strip()]
            except:
                _supported_devices = _broad_supported_devices
    return iter(_supported_devices)


class multidevices:
    __default__: Callable[..., Any]
    def __new__(cls, function: Callable[..., Any]):
        instance = super().__new__(cls)
        instance.__default__ = function
        instance.__functions = {}
        return instance
    
    @property
    def __code__(self):
        return self.__default__.__code__
    
    def __getitem__(self, device: str):
        return self.get(device)
    
    def __call__(self, *args, **kwargs):
        return self.__default__(*args, **kwargs)
    
    def get(self, device: str):
        return self.__functions.get(device, self.__default__)
    
    @overload
    def register_device(self, device: str, function: Callable[..., Any]):
        """It is used to register the function as the given device.
        """
        ...
    
    @overload
    def register_device(self, device: str):
        """It is a decorator for registering the function as the given device.
        """
        ...
    
    def register_device(self, device: str, function: Optional[Callable[..., Any]] = None):
        # Decorator Mode
        if function is None:
            def registration(function: Callable[..., Any]):
                return self.__registration(device, function)
            return registration
        elif callable(function):
            return self.__registration(device, function)
    
    def __registration(self, device: str, function: Callable[..., Any]):
        try:
            if getattr(torch, device).is_available():
                self.__functions[device] = function
                function.__manager__ = self
        except: ...
        return function


from torch import *
from . import nn
from . import image
from . import patchwork
from .functional import (
    buffer,
    is_int,
    is_float,
    shift,
    shift_,
    promote_types,
    min,
    max,
    map_range,
    map_ranges,
    linspace,
    linspace_at,
    gamma,
    gamma_div,
    invert,
    advanced_indexing,
    grow,
    apply_from_dim
)


from typing import Union, Sequence, Optional, Callable
import torch
from torch import Tensor
from .patches import unfold_space


def pool(
    input: Tensor,
    kernel_size: Union[int, Sequence[int]],
    stride: Union[int, Sequence[int]],
    padding: Union[int, Sequence[int]] = 0,
    dilation: Union[int, Sequence[int]] = 1,
    ceil_mode: bool = False,
    *,
    default_space: Union[float, str] = 0,
    reducer: Callable[[Tensor, tuple[int, ...]], Tensor]
) -> Tensor:
    spatial_ndim = input.ndim - 2
    output = unfold_space(input, kernel_size, stride, padding, dilation, ceil_mode, default_space=default_space)
    return reducer(output, tuple(range(-spatial_ndim, 0, 1)))


def max_pool(
    input: Tensor,
    kernel_size: Union[int, Sequence[int]],
    stride: Union[int, Sequence[int]],
    padding: Union[int, Sequence[int]] = 0,
    dilation: Union[int, Sequence[int]] = 1,
    ceil_mode: bool = False
) -> Tensor:
    spatial_ndim = input.ndim - 2
    # import torch.nn.functional as F
    # if spatial_ndim == 2:
    #     return F.max_pool2d(input, kernel_size, stride, padding, dilation, ceil_mode)
    # elif spatial_ndim == 3:
    #     return F.max_pool3d(input, kernel_size, stride, padding, dilation, ceil_mode)
    # elif spatial_ndim == 1:
    #     return F.max_pool1d(input, kernel_size, stride, padding, dilation, ceil_mode)
    output = unfold_space(input, kernel_size, stride, padding, dilation, ceil_mode, default_space=float("-inf"))
    return output.amax(tuple(range(-spatial_ndim, 0, 1)))


def min_pool(
    input: Tensor,
    kernel_size: Union[int, Sequence[int]],
    stride: Union[int, Sequence[int]],
    padding: Union[int, Sequence[int]] = 0,
    dilation: Union[int, Sequence[int]] = 1,
    ceil_mode: bool = False
) -> Tensor:
    spatial_ndim = input.ndim - 2
    output = unfold_space(input, kernel_size, stride, padding, dilation, ceil_mode, default_space=float("inf"))
    return output.amin(tuple(range(-spatial_ndim, 0, 1)))


def avg_pool(
    input: Tensor,
    kernel_size: Union[int, Sequence[int]],
    stride: Union[int, Sequence[int]],
    padding: Union[int, Sequence[int]] = 0,
    dilation: Union[int, Sequence[int]] = 1,
    ceil_mode: bool = False,
    count_include_pad: bool = True,
    divisor_override: Optional[int] = None
) -> Tensor:
    spatial_ndim = input.ndim - 2
    # if dilation == 1 or (len(dilation) == spatial_ndim and all(value == 1 for value in dilation)):
    #     import torch.nn.functional as F
    #     if spatial_ndim == 2:
    #         return F.avg_pool2d(input, kernel_size, stride, padding, ceil_mode, count_include_pad, divisor_override)
    #     elif spatial_ndim == 3:
    #         return F.avg_pool3d(input, kernel_size, stride, padding, ceil_mode, count_include_pad, divisor_override)
    #     elif spatial_ndim == 1 and divisor_override is None:
    #         return F.avg_pool1d(input, kernel_size, stride, padding, ceil_mode, count_include_pad)
    dim = tuple(range(-spatial_ndim, 0, 1))
    output = unfold_space(input, kernel_size, stride, padding, dilation, ceil_mode)
    if divisor_override is None:
        if count_include_pad:
            return output.mean(dim)
        else:
            output = unfold_space(input, kernel_size, stride, padding, dilation, ceil_mode)
            mask = unfold_space(
                torch.ones((1,) * input.ndim, device=input.device, dtype=torch.bool).expand(input.size(0), 1, *input.shape[2:]),
                kernel_size, stride, padding, dilation, ceil_mode
            )
            return torch.masked.mean(output, dim, mask=mask)
    else:
        return output.sum(dim) / divisor_override


import torch
from torch import Tensor
import torch.nn.functional as F
from typing import Union, Sequence


def convert_to_floating_image(image: Tensor):
    if image.is_floating_point():
        return image
    else:
        dtype_max = torch.iinfo(image.dtype).max
        image = image / dtype_max
        return image


def sobel_edges(image: Tensor):
    """Returns a tensor holding Sobel edge maps.
    
    Args:
        image (Tensor): Image tensor with shape [batch_size, channels, height, width], expected a floating point type.
    
    Returns:
        out (Tensor): Tensor holding edge maps for each channel. Returns a tensor with shape
            [batch_size, channels, height, width, 2] where the last dimension holds [sobel_edge_y, sobel_edge_x].
    """
    # Sobel Filters
    kernels = torch.tensor(
        [[[-1, -2, -1], [0, 0, 0], [1, 2, 1]], [[-1, 0, 1], [-2, 0, 2], [-1, 0, 1]]],
        dtype=image.dtype,
        device=image.device
    )
    kernels = kernels[:, None, :, :].repeat(image.size(1), 1, 1, 1)
    
    padded_image = F.pad(image, [1, 1, 1, 1], mode='reflect')
    output = F.conv2d(padded_image, kernels, groups=image.size(1)).view(image.size(0), image.size(1), -1, image.size(2), image.size(3)).permute(0, 1, 3, 4, 2)
    return output


def shift_image(input: torch.Tensor, shift: Sequence[int], fill_value = 0) -> Tensor:
    empty = torch.full_like(input, fill_value=fill_value)
    slice_none = slice(None)
    shift_slices = (..., *(slice(current_shift, None, None) if current_shift > 0 else slice(None, current_shift, None) if current_shift < 0 else slice_none for current_shift in shift))
    container_slices = (..., *(slice(None, -current_shift, None) if current_shift > 0 else slice(-current_shift, None, None) if current_shift < 0 else slice_none for current_shift in shift))
    empty[container_slices] = input[shift_slices]
    return empty


from typing import Union, Sequence, Tuple, Optional, Callable
from functools import partial
from numbers import Number
import torch
from torch import Tensor
import torch.nn.functional as F
from .patches import unfold_space, fold_space, fold_stack, patches_column, _spatialize_tuple


def conv_patches(input: Tensor, weight: Tensor, groups: int = 1) -> Tensor:
    """
    Args:
        input (Tensor): [B, C_in, out_L_0, out_L_1, ..., out_L_n, k_L_0, k_L_1, ..., k_L_n]
        weight (Tensor): [B (1), C_out, C_in // groups, k_L_0, k_L_1, ..., k_L_n]
    
    Returns:
        output (Tensor): [B, C_out, out_L_0, out_L_1, ..., out_L_n]
    
    """
    column = patches_column(input) # [B, C_in * K, L]
    column = column.view(column.size(0), groups, -1, column.size(-1)) # [B, G, C_in // G * K, L]
    flat_weight = weight.flatten(2) # [B, C_out, C_in // G * K]
    flat_weight = flat_weight.view(flat_weight.size(0), groups, -1, flat_weight.size(-1)) # [B, G, C_out // G, C_in // G * K]
    output = torch.matmul(flat_weight, column).view(input.size(0), -1, *input.shape[2:weight.ndim - 1]) # [B, G, C_out // G, L] -> [B, C_out, out_L_0, out_L_1, ..., out_L_n]
    return output


def conv(
    input: Tensor,
    weight: Tensor,
    bias: Optional[Tensor] = None,
    stride: Union[int, Sequence[int]] = 1,
    padding: Union[int, Sequence[int]] = 0,
    dilation: Union[int, Sequence[int]] = 1,
    groups: int = 1
) -> Tensor:
    """
    Args:
        input (Tensor): [B, C_in, L_0, L_1, ..., L_n]
        weight (Tensor): [C_out, C_in // groups, k_L_0, k_L_1, ..., k_L_n]
        bias (Tensor): [C_out]
    
    Returns:
        output (Tensor): [B, C_out, out_L_0, out_L_1, ..., out_L_n]
    
    """
    # spatial_ndim = input.ndim - 2
    # if spatial_ndim == 2:
    #     return F.conv2d(input, weight, bias, stride, padding, dilation, groups)
    # elif spatial_ndim == 3:
    #     return F.conv3d(input, weight, bias, stride, padding, dilation, groups)
    # elif spatial_ndim == 1:
    #     return F.conv1d(input, weight, bias, stride, padding, dilation, groups)
    return batch_conv(
        input,
        weight.unsqueeze(0),
        bias,
        stride,
        padding,
        dilation,
        groups
    )


def batch_conv(
    input: Tensor,
    weight: Tensor,
    bias: Optional[Tensor] = None,
    stride: Union[int, Sequence[int]] = 1,
    padding: Union[int, Sequence[int]] = 0,
    dilation: Union[int, Sequence[int]] = 1,
    groups: int = 1
) -> Tensor:
    """
    Args:
        input (Tensor): [B, C_in, L_0, L_1, ..., L_n]
        weight (Tensor): [B, C_out, C_in // groups, k_L_0, k_L_1, ..., k_L_n]
        bias (Tensor): [B, C_out]
    
    Returns:
        output (Tensor): [B, C_out, out_L_0, out_L_1, ..., out_L_n]
    
    """
    input = unfold_space(input, weight.shape[3:], stride, padding, dilation)
    output = conv_patches(input, weight, groups)
    if bias is not None:
        output = output + bias[(..., *((None,) * (input.ndim - 2)))] # [B, C_out, out_L_0, out_L_1, ..., out_L_n]
    return output


def conv_patches_explicitly(input: Tensor, weight: Tensor, groups: int = 1):
    """
    Args:
        input (Tensor): [B, C_in, out_L_0, out_L_1, ..., out_L_n, k_L_0, k_L_1, ..., k_L_n]
        weight (Tensor): [B (1), C_out, C_in // groups, k_L_0, k_L_1, ..., k_L_n]
    
    Returns:
        output (Tensor): [B, C_out, out_L_0, out_L_1, ..., out_L_n]
    
    """
    spatial_ndim = weight.ndim - 3
    input = input.view(input.size(0), groups, -1, *input.shape[2:]) # [B, G, C_in // G, out_L_0, out_L_1, ..., out_L_n, k_L_0, k_L_1, ..., k_L_n]
    input = input.permute(0, 1, *range(3, weight.ndim), 2, *range(weight.ndim, input.ndim))[:, :, None] # [B, G, 1, out_L_0, out_L_1, ..., out_L_n, C_in // G, k_L_0, k_L_1, ..., k_L_n]
    weight = weight.view(weight.size(0), groups, -1, *((1,) * spatial_ndim), *weight.shape[2:]) # [B, G, C_out // G, 1 (0), 1 (1), ..., 1 (n), C_in // G, k_L_0, k_L_1, ..., k_L_n]
    output = torch.sum(input * weight, tuple(range(-spatial_ndim - 1, 0, 1))).flatten(1, 2) # [B, C_out, out_L_0, out_L_1, ..., out_L_n]
    return output


def conv_explicitly(
    input: Tensor,
    weight: Tensor,
    bias: Optional[Tensor] = None,
    stride: Union[int, Sequence[int]] = 1,
    padding: Union[int, Sequence[int]] = 0,
    dilation: Union[int, Sequence[int]] = 1,
    groups: int = 1
) -> Tensor:
    """
    Args:
        input (Tensor): [B, C_in, L_0, L_1, ..., L_n]
        weight (Tensor): [C_out, C_in // groups, k_L_0, k_L_1, ..., k_L_n]
        bias (Tensor): [C_out]
    
    Returns:
        output (Tensor): [B, C_out, out_L_0, out_L_1, ..., out_L_n]
    
    """
    return batch_conv_explicitly(
        input,
        weight.unsqueeze(0),
        bias,
        stride,
        padding,
        dilation,
        groups
    )


def batch_conv_explicitly(
    input: Tensor,
    weight: Tensor,
    bias: Optional[Tensor] = None,
    stride: Union[int, Sequence[int]] = 1,
    padding: Union[int, Sequence[int]] = 0,
    dilation: Union[int, Sequence[int]] = 1,
    groups: int = 1
) -> Tensor:
    """
    Args:
        input (Tensor): [B, C_in, L_0, L_1, ..., L_n]
        weight (Tensor): [B (1), C_out, C_in // groups, k_L_0, k_L_1, ..., k_L_n]
        bias (Tensor): [B (1), C_out]
    
    Returns:
        output (Tensor): [B, C_out, out_L_0, out_L_1, ..., out_L_n]
    
    """
    input = unfold_space(input, weight.shape[3:], stride, padding, dilation)
    output = conv_patches_explicitly(input, weight, groups)
    if bias is not None:
        output = output + bias[(..., *((None,) * (input.ndim - 2)))] # [B, C_out, out_L_0, out_L_1, ..., out_L_n]
    return output


def masked_conv_patches(
    input: Tensor,
    weight: Tensor,
    groups: int = 1,
    *,
    input_mask: Optional[Tensor] = None,
    weight_mask: Optional[Tensor] = None,
    soft_mode: bool = False
):
    """
    Args:
        input (Tensor): [B, C_in, out_L_0, out_L_1, ..., out_L_n, k_L_0, k_L_1, ..., k_L_n]
        weight (Tensor): [B (1), C_out, C_in // groups, k_L_0, k_L_1, ..., k_L_n]
        input_mask (Tensor): [B (1), C_in (1), out_L_0, out_L_1, ..., out_L_n, k_L_0, k_L_1, ..., k_L_n]
        weight_mask: [B (1), C_out (1), (C_in // groups) (1), k_L_0, k_L_1, ..., k_L_n]
    
    Returns:
        output (Tensor): [B, C_out, out_L_0, out_L_1, ..., out_L_n]
    
    """
    spatial_ndim = weight.ndim - 3
    mask = None
    
    input_shape = input.shape
    new_input_shape = (input.size(0), groups, -1, *input.shape[2:]) # [B, G, C_in // G, out_L_0, out_L_1, ..., out_L_n, k_L_0, k_L_1, ..., k_L_n]
    input = input.view(new_input_shape)
    input_permutation = (0, 1, *range(3, weight.ndim), 2, *range(weight.ndim, input.ndim)) # [B, G, out_L_0, out_L_1, ..., out_L_n, C_in // G, k_L_0, k_L_1, ..., k_L_n]
    input = input.permute(input_permutation)[:, :, None] # [B, G, 1, out_L_0, out_L_1, ..., out_L_n, C_in // G, k_L_0, k_L_1, ..., k_L_n]
    
    weight_shape = weight.shape
    new_weight_shape = (weight.size(0), groups, -1, *((1,) * spatial_ndim), *weight.shape[2:]) # [B, G, C_out // G, 1 (0), 1 (1), ..., 1 (n), C_in // G, k_L_0, k_L_1, ..., k_L_n]
    weight = weight.view(new_weight_shape)
    
    weighted_input = input * weight
    
    if input_mask is not None:
        input_mask = input_mask.expand(input_shape).view(new_input_shape).permute(input_permutation)[:, :, None]
        if not soft_mode:
            input_mask = input_mask.amin(tuple(range(-spatial_ndim, 0, 1)), keepdim=True).expand(input_mask.shape)
        mask = input_mask
    
    if weight_mask is not None:
        weight_mask = weight_mask.expand(weight_shape).view(new_weight_shape)
        weighted_input_shape = weighted_input.shape
        if mask is None:
            mask = weight_mask.expand(weighted_input_shape)
        else:
            mask = (mask & weight_mask).expand(weighted_input_shape)
    
    output = torch.masked.sum(weighted_input, tuple(range(-spatial_ndim - 1, 0, 1)), mask=mask).flatten(1, 2) # [B, C_out, out_L_0, out_L_1, ..., out_L_n]
    return output


def masked_conv(
    input: Tensor,
    weight: Tensor,
    bias: Optional[Tensor] = None,
    stride: Union[int, Sequence[int]] = 1,
    padding: Union[int, Sequence[int]] = 0,
    dilation: Union[int, Sequence[int]] = 1,
    groups: int = 1,
    *,
    input_mask: Optional[Tensor] = None,
    weight_mask: Optional[Tensor] = None,
    soft_mode: bool = False
) -> Tensor:
    """
    Args:
        input (Tensor): [B, C_in, L_0, L_1, ..., L_n]
        weight (Tensor): [B (1), C_out, C_in // groups, k_L_0, k_L_1, ..., k_L_n]
        bias (Tensor): [B (1), C_out]
        input_mask (Tensor): [B (1), C_in (1), L_0, L_1, ..., L_n]
        weight_mask: [B (1), C_out (1), (C_in // groups) (1), k_L_0, k_L_1, ..., k_L_n]
    
    Returns:
        output (Tensor): [B, C_out, out_L_0, out_L_1, ..., out_L_n]
    
    """
    kernel_size = weight.shape[3:]
    input = unfold_space(input, kernel_size, stride, padding, dilation)
    if input_mask is None:
        input_mask = None
    else:
        input_mask = unfold_space(input_mask, kernel_size, stride, padding, dilation)
    output = masked_conv_patches(input, weight, groups, input_mask=input_mask, weight_mask=weight_mask, soft_mode=soft_mode)
    if bias is not None:
        output = output + bias[(..., *((None,) * (input.ndim - 2)))] # [B, C_out, out_L_0, out_L_1, ..., out_L_n]
    return output


def conv_transpose_patches(input: Tensor, weight: Tensor, groups: int = 1) -> Tensor:
    """
    Args:
        input (Tensor): [B, C_in, L_0, L_1, ..., L_n]
        weight (Tensor): [B (1), C_in, C_out // groups, k_L_0, k_L_1, ..., k_L_n]
    
    Returns:
        output (Tensor): [B, C_out, L_0, L_1, ..., L_n, k_L_0, k_L_1, ..., k_L_n]
    
    """
    input_ndim = input.ndim
    
    batch_size = input.size(0)
    flat_input = input.flatten(2) # [B, C_in, L]
    reshaped_input = flat_input.view(batch_size, groups, -1, flat_input.size(-1)).permute(0, 1, 3, 2) # [B, G, L, C_in // G]
    flat_weight = weight.flatten(2) # [B, C_in, (C_out // G) * K]
    reshaped_weight = flat_weight.view(weight.size(0), groups, -1, flat_weight.size(-1)) # [B, G, C_in // G, (C_out // G) * K]

    # The physical meaning calculated in this step is that for each pixel position (L) of the input,
    # a local block ((C_out // G) * K) will be generated.
    output = torch.matmul(reshaped_input, reshaped_weight) # [B, G, L, (C_out // G) * K]
    output = output.view(batch_size, groups, *input.shape[2:], -1, *weight.shape[3:]) # [B, G, L_0, L_1, ..., L_n, C_out // G, k_L_0, k_L_1, ..., k_L_n]
    output = output.permute(0, 1, input_ndim, *range(2, input_ndim), *range(input_ndim + 1, output.ndim)) # [B, G, C_out // G, L_0, L_1, ..., L_n, k_L_0, k_L_1, ..., k_L_n]
    output = output.reshape(batch_size, -1, *input.shape[2:], *weight.shape[3:]) # [B, C_out, L_0, L_1, ..., L_n, k_L_0, k_L_1, ..., k_L_n]
    
    return output


def conv_transpose(
    input: Tensor,
    weight: Tensor,
    bias: Optional[Tensor] = None,
    stride: Union[int, Sequence[int]] = 1,
    padding: Union[int, Sequence[int]] = 0,
    output_padding: Union[int, Sequence[int]] = 0,
    groups: int = 1,
    dilation: Union[int, Sequence[int]] = 1,
    *,
    default_space: Union[float, str] = 0
) -> Tensor:
    """
    Args:
        input (Tensor): [B, C_in, L_0, L_1, ..., L_n]
        weight (Tensor): [C_in, C_out // groups, k_L_0, k_L_1, ..., k_L_n]
    
    Returns:
        output (Tensor): [B, C_out, L_0, L_1, ..., L_n]
    
    """
    # spatial_ndim = input.ndim - 2
    # if spatial_ndim == 2:
    #     return F.conv_transpose2d(
    #         input,
    #         bias,
    #         stride,
    #         padding,
    #         output_padding,
    #         groups,
    #         dilation
    #     )
    # elif spatial_ndim == 3:
    #     return F.conv_transpose3d(
    #         input,
    #         bias,
    #         stride,
    #         padding,
    #         output_padding,
    #         groups,
    #         dilation
    #     )
    # elif spatial_ndim == 1:
    #     return F.conv_transpose1d(
    #         input,
    #         bias,
    #         stride,
    #         padding,
    #         output_padding,
    #         groups,
    #         dilation
    #     )
    return batch_conv_transpose(
        input,
        weight.unsqueeze(0),
        bias,
        stride,
        padding,
        output_padding,
        groups,
        dilation,
        default_space=default_space
    )


def batch_conv_transpose(
    input: Tensor,
    weight: Tensor,
    bias: Optional[Tensor] = None,
    stride: Union[int, Sequence[int]] = 1,
    padding: Union[int, Sequence[int]] = 0,
    output_padding: Union[int, Sequence[int]] = 0,
    groups: int = 1,
    dilation: Union[int, Sequence[int]] = 1,
    *,
    default_space: Union[float, str] = 0
) -> Tensor:
    """
    Args:
        input (Tensor): [B, C_in, L_0, L_1, ..., L_n]
        weight (Tensor): [B (1), C_in, C_out // groups, k_L_0, k_L_1, ..., k_L_n]
    
    Returns:
        output (Tensor): [B, C_out, L_0, L_1, ..., L_n]
    
    """
    return _conv_transpose_template(
        input,
        weight,
        bias,
        stride,
        padding,
        output_padding,
        groups,
        dilation,
        default_space=default_space,
        expander=conv_transpose_patches,
        folder=fold_stack
    )


def deconv(
    input: Tensor,
    weight: Tensor,
    bias: Optional[Tensor] = None,
    stride: Union[int, Sequence[int]] = 1,
    padding: Union[int, Sequence[int]] = 0,
    output_padding: Union[int, Sequence[int]] = 0,
    groups: int = 1,
    dilation: Union[int, Sequence[int]] = 1,
    *,
    default_space: Union[float, str] = 0
) -> Tensor:
    """
    Args:
        input (Tensor): [B, C_in, L_0, L_1, ..., L_n]
        weight (Tensor): [C_in, C_out // groups, k_L_0, k_L_1, ..., k_L_n]
    
    Returns:
        output (Tensor): [B, C_out, L_0, L_1, ..., L_n]
    
    """
    return batch_deconv(
        input,
        weight.unsqueeze(0),
        bias,
        stride,
        padding,
        output_padding,
        groups,
        dilation,
        default_space=default_space
    )


def batch_deconv(
    input: Tensor,
    weight: Tensor,
    bias: Optional[Tensor] = None,
    stride: Union[int, Sequence[int]] = 1,
    padding: Union[int, Sequence[int]] = 0,
    output_padding: Union[int, Sequence[int]] = 0,
    groups: int = 1,
    dilation: Union[int, Sequence[int]] = 1,
    *,
    default_space: Union[float, str] = 0
) -> Tensor:
    """
    Args:
        input (Tensor): [B, C_in, L_0, L_1, ..., L_n]
        weight (Tensor): [B (1), C_in, C_out // groups, k_L_0, k_L_1, ..., k_L_n]
    
    Returns:
        output (Tensor): [B, C_out, L_0, L_1, ..., L_n]
    
    """
    return _conv_transpose_template(
        input,
        weight,
        bias,
        stride,
        padding,
        output_padding,
        groups,
        dilation,
        default_space=default_space,
        expander=conv_transpose_patches,
        folder=fold_space
    )


def conv_transpose_patches_explicitly(input: Tensor, weight: Tensor, groups: int = 1) -> Tensor:
    """
    Args:
        input (Tensor): [B, C_in, L_0, L_1, ..., L_n]
        weight (Tensor): [B (1), C_in, C_out // groups, k_L_0, k_L_1, ..., k_L_n]
    
    Returns:
        output (Tensor): [B, C_out, L_0, L_1, ..., L_n, k_L_0, k_L_1, ..., k_L_n]
    
    """
    spatial_ndim = input.ndim - 2
    
    reshaped_input = input.view(input.size(0), groups, -1, *input.shape[2:]) # [B, G, C_in // G, L_0, L_1, ..., L_n]
    reshaped_input = reshaped_input.permute(0, 1, *range(3, reshaped_input.ndim), 2) # [B, G, L_0, L_1, ..., L_n, C_in // G]
    reshaped_input = reshaped_input.unsqueeze(2)[(..., ) + (None,) * spatial_ndim] # [B, G, 1, L_0, L_1, ..., L_n, C_in // G, 1 (0), 1 (1), ..., 1 (n)]
    reshaped_weight = weight.view(weight.size(0), groups, -1, weight.size(2), *weight.shape[3:]) # [B, G, C_in // G, C_out // G, k_L_0, k_L_1, ..., k_L_n]
    reshaped_weight = reshaped_weight.permute(0, 1, 3, 2, *range(4, reshaped_weight.ndim)) # [B, G, C_out // G, C_in // G, k_L_0, k_L_1, ..., k_L_n]
    reshaped_weight = reshaped_weight[(*(slice(None),) * 3, *((None,) * spatial_ndim))] # [B, G, C_out // G, 1 (0), 1 (1), ..., 1 (n), C_in // G, k_L_0, k_L_1, ..., k_L_n]
    
    output = torch.sum(reshaped_input * reshaped_weight, dim=-spatial_ndim - 1).flatten(1, 2) # [B, C_out, L_0, L_1, ..., L_n, k_L_0, k_L_1, ..., k_L_n]
    return output


def conv_transpose_explicitly(
    input: Tensor,
    weight: Tensor,
    bias: Optional[Tensor] = None,
    stride: Union[int, Sequence[int]] = 1,
    padding: Union[int, Sequence[int]] = 0,
    output_padding: Union[int, Sequence[int]] = 0,
    groups: int = 1,
    dilation: Union[int, Sequence[int]] = 1,
    *,
    default_space: Union[float, str] = 0
) -> Tensor:
    """
    Args:
        input (Tensor): [B, C_in, L_0, L_1, ..., L_n]
        weight (Tensor): [C_in, C_out // groups, k_L_0, k_L_1, ..., k_L_n]
    
    Returns:
        output (Tensor): [B, C_out, L_0, L_1, ..., L_n]
    
    """
    return batch_conv_transpose_explicitly(
        input,
        weight.unsqueeze(0),
        bias,
        stride,
        padding,
        output_padding,
        groups,
        dilation,
        default_space=default_space
    )


def batch_conv_transpose_explicitly(
    input: Tensor,
    weight: Tensor,
    bias: Optional[Tensor] = None,
    stride: Union[int, Sequence[int]] = 1,
    padding: Union[int, Sequence[int]] = 0,
    output_padding: Union[int, Sequence[int]] = 0,
    groups: int = 1,
    dilation: Union[int, Sequence[int]] = 1,
    *,
    default_space: Union[float, str] = 0
) -> Tensor:
    """
    Args:
        input (Tensor): [B, C_in, L_0, L_1, ..., L_n]
        weight (Tensor): [B (1), C_in, C_out // groups, k_L_0, k_L_1, ..., k_L_n]
    
    Returns:
        output (Tensor): [B, C_out, L_0, L_1, ..., L_n]
    
    """
    return _conv_transpose_template(
        input,
        weight,
        bias,
        stride,
        padding,
        output_padding,
        groups,
        dilation,
        default_space=default_space,
        expander=conv_transpose_patches_explicitly,
        folder=fold_stack
    )


def deconv_explicitly(
    input: Tensor,
    weight: Tensor,
    bias: Optional[Tensor] = None,
    stride: Union[int, Sequence[int]] = 1,
    padding: Union[int, Sequence[int]] = 0,
    output_padding: Union[int, Sequence[int]] = 0,
    groups: int = 1,
    dilation: Union[int, Sequence[int]] = 1,
    *,
    default_space: Union[float, str] = 0
) -> Tensor:
    """
    Args:
        input (Tensor): [B, C_in, L_0, L_1, ..., L_n]
        weight (Tensor): [C_in, C_out // groups, k_L_0, k_L_1, ..., k_L_n]
    
    Returns:
        output (Tensor): [B, C_out, L_0, L_1, ..., L_n]
    
    """
    return batch_deconv_explicitly(
        input,
        weight.unsqueeze(0),
        bias,
        stride,
        padding,
        output_padding,
        groups,
        dilation,
        default_space=default_space
    )


def batch_deconv_explicitly(
    input: Tensor,
    weight: Tensor,
    bias: Optional[Tensor] = None,
    stride: Union[int, Sequence[int]] = 1,
    padding: Union[int, Sequence[int]] = 0,
    output_padding: Union[int, Sequence[int]] = 0,
    groups: int = 1,
    dilation: Union[int, Sequence[int]] = 1,
    *,
    default_space: Union[float, str] = 0
) -> Tensor:
    """
    Args:
        input (Tensor): [B, C_in, L_0, L_1, ..., L_n]
        weight (Tensor): [B (1), C_in, C_out // groups, k_L_0, k_L_1, ..., k_L_n]
    
    Returns:
        output (Tensor): [B, C_out, L_0, L_1, ..., L_n]
    
    """
    return _conv_transpose_template(
        input,
        weight,
        bias,
        stride,
        padding,
        output_padding,
        groups,
        dilation,
        default_space=default_space,
        expander=conv_transpose_patches_explicitly,
        folder=fold_space
    )


def masked_conv_transpose_patches(
    input: Tensor,
    weight: Tensor,
    groups: int = 1,
    *,
    input_mask: Optional[Tensor] = None,
    weight_mask: Optional[Tensor] = None,
    soft_mode: bool = False
) -> Tensor:
    """
    Args:
        input (Tensor): [B, C_in, L_0, L_1, ..., L_n]
        weight (Tensor): [B (1), C_in, C_out // groups, k_L_0, k_L_1, ..., k_L_n]
    
    Returns:
        output (Tensor): [B, C_out, L_0, L_1, ..., L_n, k_L_0, k_L_1, ..., k_L_n]
    
    """
    spatial_ndim = input.ndim - 2
    mask = None
    
    input_shape = input.shape
    new_input_shape = (input.size(0), groups, -1, *input.shape[2:]) # [B, G, C_in // G, L_0, L_1, ..., L_n]
    reshaped_input = input.view(new_input_shape)
    input_permutation = (0, 1, *range(3, reshaped_input.ndim), 2) # [B, G, L_0, L_1, ..., L_n, C_in // G]
    reshaped_input = reshaped_input.permute(input_permutation)
    new_input_slice = (..., ) + (None,) * spatial_ndim
    reshaped_input = reshaped_input.unsqueeze(2)[new_input_slice] # [B, G, 1, L_0, L_1, ..., L_n, C_in // G, 1 (0), 1 (1), ..., 1 (n)]
    
    weight_shape = weight.shape
    new_weight_shape = (weight.size(0), groups, -1, weight.size(2), *weight.shape[3:]) # [B, G, C_in // G, C_out // G, k_L_0, k_L_1, ..., k_L_n]
    reshaped_weight = weight.view(new_weight_shape)
    weight_permutation = (0, 1, 3, 2, *range(4, reshaped_weight.ndim)) # [B, G, C_out // G, C_in // G, k_L_0, k_L_1, ..., k_L_n]
    reshaped_weight = reshaped_weight.permute(weight_permutation)
    new_weight_slice = (*(slice(None),) * 3, *((None,) * spatial_ndim))
    reshaped_weight = reshaped_weight[new_weight_slice] # [B, G, C_out // G, 1 (0), 1 (1), ..., 1 (n), C_in // G, k_L_0, k_L_1, ..., k_L_n]
    
    weighted_input = input * weight
    
    if input_mask is not None:
        input_mask = input_mask.expand(input_shape).view(new_input_shape).permute(input_permutation).unsqueeze(2)[new_input_slice]
        if not soft_mode:
            input_mask = input_mask.amin(tuple(range(-spatial_ndim, 0, 1)), keepdim=True).expand(input_mask.shape)
        mask = input_mask
    
    if weight_mask is not None:
        weight_mask = weight_mask.expand(weight_shape).view(new_weight_shape).permute(weight_permutation)[new_weight_slice]
        weighted_input_shape = weighted_input.shape
        if mask is None:
            mask = weight_mask.expand(weighted_input_shape)
        else:
            mask = (mask & weight_mask).expand(weighted_input_shape)
    
    output = torch.masked.sum(reshaped_input * reshaped_weight, dim=-spatial_ndim - 1, mask=mask).flatten(1, 2) # [B, C_out, L_0, L_1, ..., L_n, k_L_0, k_L_1, ..., k_L_n]
    return output


def masked_conv_transpose(
    input: Tensor,
    weight: Tensor,
    bias: Optional[Tensor] = None,
    stride: Union[int, Sequence[int]] = 1,
    padding: Union[int, Sequence[int]] = 0,
    output_padding: Union[int, Sequence[int]] = 0,
    groups: int = 1,
    dilation: Union[int, Sequence[int]] = 1,
    *,
    default_space: Union[float, str] = 0,
    input_mask: Optional[Tensor] = None,
    weight_mask: Optional[Tensor] = None,
    soft_mode: bool = False
) -> Tensor:
    """
    Args:
        input (Tensor): [B, C_in, L_0, L_1, ..., L_n]
        weight (Tensor): [B (1), C_in, C_out // groups, k_L_0, k_L_1, ..., k_L_n]
    
    Returns:
        output (Tensor): [B, C_out, L_0, L_1, ..., L_n]
    
    """
    return _conv_transpose_template(
        input,
        weight,
        bias,
        stride,
        padding,
        output_padding,
        groups,
        dilation,
        default_space=default_space,
        expander=masked_conv_transpose_patches,
        folder=fold_stack,
        input_mask=input_mask,
        weight_mask=weight_mask,
        soft_mode=soft_mode
    )


def masked_deconv(
    input: Tensor,
    weight: Tensor,
    bias: Optional[Tensor] = None,
    stride: Union[int, Sequence[int]] = 1,
    padding: Union[int, Sequence[int]] = 0,
    output_padding: Union[int, Sequence[int]] = 0,
    groups: int = 1,
    dilation: Union[int, Sequence[int]] = 1,
    *,
    default_space: Union[float, str] = 0,
    input_mask: Optional[Tensor] = None,
    weight_mask: Optional[Tensor] = None,
    soft_mode: bool = False
) -> Tensor:
    """
    Args:
        input (Tensor): [B, C_in, L_0, L_1, ..., L_n]
        weight (Tensor): [B (1), C_in, C_out // groups, k_L_0, k_L_1, ..., k_L_n]
    
    Returns:
        output (Tensor): [B, C_out, L_0, L_1, ..., L_n]
    
    """
    return _conv_transpose_template(
        input,
        weight,
        bias,
        stride,
        padding,
        output_padding,
        groups,
        dilation,
        default_space=default_space,
        expander=masked_conv_transpose_patches,
        folder=fold_space,
        input_mask=input_mask,
        weight_mask=weight_mask,
        soft_mode=soft_mode
    )


def multisize_conv(
    input: Tensor,
    weight: Tensor,
    bias: Optional[Tensor] = None,
    stride: Union[int, Sequence[int]] = 1,
    padding: Union[int, Sequence[int]] = 0,
    dilation: Union[int, Sequence[int]] = 1,
    groups: int = 1,
    *,
    weight_box: Tensor
) -> Tensor:
    """This function has not undergone rigorous testing and may change at any time.
    
    Args:
        input (Tensor): [B, C_in, L_0, L_1, ..., L_n]
        weight (Tensor): [B (1), C_out, C_in // groups, k_L_0, k_L_1, ..., k_L_n]
        bias (Tensor): [B (1), C_out]
        weight_box (Tensor): [B (1), C_out (1), C_in // groups (1), 2, n]
    
    Returns:
        output (Tensor): [B, C_out, out_L_0, out_L_1, ..., out_L_n]
    
    """
    input, input_mask, weight, weight_mask = _multisize_kernel_preprocess(input, weight, weight_box, stride, padding, dilation)
    
    kernel_size = weight.shape[3:]
    input = unfold_space(input, kernel_size, stride, padding, dilation)
    input_mask = unfold_space(input_mask, kernel_size, stride, padding, dilation)
    output = masked_conv_patches(input, weight, groups, input_mask=input_mask, weight_mask=weight_mask)
    if bias is not None:
        output = output + bias[(..., *((None,) * (input.ndim - 2)))] # [B, C_out, out_L_0, out_L_1, ..., out_L_n]
    return output


def multisize_masked_conv(
    input: Tensor,
    weight: Tensor,
    bias: Optional[Tensor] = None,
    stride: Union[int, Sequence[int]] = 1,
    padding: Union[int, Sequence[int]] = 0,
    dilation: Union[int, Sequence[int]] = 1,
    groups: int = 1,
    *,
    weight_box: Tensor,
    input_mask: Optional[Tensor] = None,
    weight_mask: Optional[Tensor] = None
) -> Tensor:
    """This function has not undergone rigorous testing and may change at any time.
    
    Args:
        input (Tensor): [B, C_in, L_0, L_1, ..., L_n]
        weight (Tensor): [B (1), C_out, C_in // groups, k_L_0, k_L_1, ..., k_L_n]
        bias (Tensor): [B (1), C_out]
        weight_box (Tensor): [B (1), C_out (1), C_in // groups (1), 2, n]
        input_mask (Tensor): [B (1), C_in (1), L_0, L_1, ..., L_n]
        weight_mask: [B (1), C_out (1), (C_in // groups) (1), k_L_0, k_L_1, ..., k_L_n]
    
    Returns:
        output (Tensor): [B, C_out, out_L_0, out_L_1, ..., out_L_n]
    
    """
    input, content_mask, weight, box_mask = _multisize_kernel_preprocess(input, weight, weight_box, stride, padding, dilation)
    
    if input_mask is None:
        input_mask = content_mask
    else:
        input_mask = input_mask & content_mask
    if weight_mask is None:
        weight_mask = box_mask
    else:
        weight_mask = weight_mask & box_mask
    
    kernel_size = weight.shape[3:]
    input = unfold_space(input, kernel_size, stride, padding, dilation)
    input_mask = unfold_space(input_mask, kernel_size, stride, padding, dilation)
    output = masked_conv_patches(input, weight, groups, input_mask=input_mask, weight_mask=weight_mask)
    if bias is not None:
        output = output + bias[(..., *((None,) * (input.ndim - 2)))] # [B, C_out, out_L_0, out_L_1, ..., out_L_n]
    return output


def _conv_transpose_template(
    input: Tensor,
    weight: Tensor,
    bias: Optional[Tensor] = None,
    stride: Union[int, Sequence[int]] = 1,
    padding: Union[int, Sequence[int]] = 0,
    output_padding: Union[int, Sequence[int]] = 0,
    groups: int = 1,
    dilation: Union[int, Sequence[int]] = 1,
    *,
    default_space: Union[float, str] = 0,
    expander: Callable[[Tensor, Tensor, int], Tensor],
    folder: Callable[[Tensor, Union[int, Sequence[int]], Union[int, Sequence[int]], Union[int, Sequence[int]]], Tensor],
    **kwargs
) -> Tensor:
    if isinstance(default_space, Number):
        functional_pad = partial(F.pad, mode='constant', value=default_space)
    else:
        functional_pad = partial(F.pad, mode=default_space)
    
    spatial_ndim = input.ndim - 2
    
    stride = _spatialize_tuple(stride, spatial_ndim)
    
    while output_padding != 0:
        if isinstance(output_padding, int):
            output_slices = (slice(None), slice(None), *(slice(output_padding - step) for step in stride))
            output_padding = [pad_value for _ in range(spatial_ndim) for pad_value in (0, 1)]
        elif any(value != 0 for value in output_padding):
            output_slices = (slice(None), slice(None), *(slice(None) if pad == 0 else slice(pad - step) for pad, step in zip(output_padding, stride)))
            output_padding = [pad_value for value in reversed(output_padding) for pad_value in (0, 1 if value > 0 else 0)]
        else:
            output_slices = (slice(None),) * input.ndim
            break
        input = functional_pad(input, output_padding)
        break
    else:
        output_slices = (slice(None),) * input.ndim
    
    patches = expander(input, weight, groups, **kwargs)
    output = folder(patches, stride, padding, dilation)[output_slices]
    
    if bias is not None:
        output = output + bias[(..., *((None,) * spatial_ndim))] # [B, C_out, out_L_0, out_L_1, ..., out_L_n]
    
    return output


def _box_length(box: torch.Tensor) -> torch.Tensor:
    return box.diff(dim=-2).squeeze_(-2).add_(1) # [B, C, N]


def _box_mask(box: Tensor, spatial_shape: Sequence[int]) -> torch.Tensor:
    batch_size, num_channels = box.size(0), box.size(1)
    spatial_ndim = box.size(-1)
    spatial_shape: Tensor = torch.as_tensor(spatial_shape, device=box.device)
    total_elements = torch.prod(spatial_shape)
    
    indices: Tensor = torch.arange(total_elements, device=box.device)
    
    # Calculate the step size for each dimension [spatial_ndim]
    reverse_cumprod = torch.cumprod(spatial_shape.flip(0), dim=0).flip(0)
    strides = torch.roll(reverse_cumprod, -1)
    strides[-1] = 1
    
    # Calculate the coordinates of each dimension [total_elements, spatial_ndim]
    indices_expanded = indices.unsqueeze(1).expand(total_elements, spatial_ndim)
    strides_expanded = strides.unsqueeze(0).expand(total_elements, spatial_ndim)
    shape_expanded = spatial_shape.unsqueeze(0).expand(total_elements, spatial_ndim)
    
    coordinates = ((indices_expanded // strides_expanded) % shape_expanded).view(*spatial_shape, spatial_ndim)
    coordinates = coordinates.view(1, 1, *spatial_shape, spatial_ndim).expand(batch_size, num_channels, -1, -1, -1)
    
    view_shape = (batch_size, num_channels, *((1,) * spatial_ndim), spatial_ndim)
    min_coordinates = box[:, :, 0, :].view(view_shape)
    max_coordinates = box[:, :, 1, :].view(view_shape)
    
    # Check if each point is located within the bounding box in all dimensions
    in_bounds = (coordinates >= min_coordinates) & (coordinates <= max_coordinates) # [B, C, L_0, L_1, ..., L_n, n]
    
    # Points must satisfy conditions in all dimensions to be within the bounding box
    mask = in_bounds.all(dim=-1).bool() # [B, C, L_0, L_1, ..., L_n]
    
    return mask


def _multisize_kernel_preprocess(
    input: Tensor,
    weight: Tensor,
    weight_box: Tensor,
    stride: Union[int, Sequence[int]] = 1,
    padding: Union[int, Sequence[int]] = 0,
    dilation: Union[int, Sequence[int]] = 1,
) -> Tuple[Tensor, Tensor, Tensor, Tensor]:
    """
    Args:
        input (Tensor): [B, C_in, L_0, L_1, ..., L_n]
        weight (Tensor): [B (1), C_out, C_in // groups, k_L_0, k_L_1, ..., k_L_n]
        weight_box (Tensor): [B (1), C_out (1), C_in // groups (1), 2, n]
    
    """
    device = weight.device

    spatial_ndim = weight_box.size(-1)
    kernel_size = _box_length(weight_box) # [B, C_out, C_in // G, n]
    box_mask = _box_mask(weight_box.view(-1, 1, *weight_box.shape[-2:]), weight.shape[3:]).view_as(weight)
    if weight_box.any(weight_box[..., 0, :] != 0):
        aligned_box = torch.empty_like(weight_box)
        aligned_box[..., 0, :].fill_(0)
        aligned_box[..., 1, :].copy_(kernel_size).sub_(1)
        aligned_box_mask = _box_mask(aligned_box.view(-1, 1, *aligned_box.shape[-2:]), weight.shape[3:])
        new_weight = torch.zeros_like(weight)
        new_weight.masked_scatter_(aligned_box_mask, weight.masked_select(box_mask))
        weight = new_weight
        box_mask = aligned_box_mask
    
    spatial_shape: Tensor = torch.as_tensor(input.shape[2:], device=device)[None, None, None, ...]
    stride: Tensor = torch.as_tensor(_spatialize_tuple(stride, spatial_ndim), device=device)[None, None, None, ...]
    padding: Tensor = torch.as_tensor(_spatialize_tuple(padding, spatial_ndim), device=device)[None, None, None, ...]
    dilation: Tensor = torch.as_tensor(_spatialize_tuple(dilation, spatial_ndim), device=device)[None, None, None, ...]
    
    nature_factor = dilation * (1 - kernel_size) - 1
    out_spatial_shape = ((spatial_shape + 2 * padding + nature_factor) / stride).floor_().to(spatial_shape.dtype) # typically requires adding 1, but not necessary here
    max_dim = tuple(range(spatial_shape.ndim - 1))
    max_spatial_shape = out_spatial_shape.amax(dim=max_dim, keepdim=True)
    reverse_full_padding = max_spatial_shape * stride - nature_factor - spatial_shape
    max_full_padding = reverse_full_padding.amax(dim=max_dim)
    squeezed_spatial_shape = spatial_shape.squeeze()
    max_out_spatial_shape = squeezed_spatial_shape + max_full_padding
    
    content_mask = torch.zeros(max_out_spatial_shape.tolist(), device=input.device, dtype=torch.bool)[None, None, ...]
    content_mask[(..., *(slice(length) for length in squeezed_spatial_shape))].fill_(1)
    input = F.pad(input, [pad_value for value in reversed(max_full_padding) for pad_value in (0, value)], mode='constant', value=0)
    
    return input, content_mask, weight, box_mask


from .patches import (
    unfold,
    unfold_space,
    fold,
    fold_roll,
    fold_space,
    fold_stack,
    patches_column,
    keep_kernel_dim,
    expand_kernel_dim
)
from .pooling import (
    pool,
    max_pool,
    min_pool,
    avg_pool
)
from .convolution import (
    conv,
    batch_conv,
    conv_patches,
    conv_explicitly,
    batch_conv_explicitly,
    conv_patches_explicitly,
    masked_conv,
    masked_conv_patches,
    conv_transpose,
    batch_conv_transpose,
    deconv,
    batch_deconv,
    conv_transpose_patches,
    batch_conv_transpose_explicitly,
    deconv_explicitly,
    batch_deconv_explicitly,
    conv_transpose_patches_explicitly,
    masked_conv_transpose,
    masked_deconv,
    masked_conv_transpose_patches
)